/*
  Filename    : SampleGrab.cpp
  Author      : roar@tordivel.no
  Description : Scorpion Camera Interface, SCI, grab standard interface implementation

	Revisions   : Version 1.0 - 08 May 2008

	General:
	  The Scorpion Camera Interface is a general interface used for image aqcuistion by
		Scorpion and ScorpionCapture. The interface has both required and optional methods
		due to complexity of the cameramodel.

		The cameradriver is responsible to maintain camera configuration as needed to secure
		proper/repeatable camera operation. Scorpion supplies a system directory at startup which
		should be used to store configuration files, see the note on "setSystemPath" method.

		This sample code uses a multimedia timer to generate images on the grab command.
		
		The sample code is 'as is' and may may be used and modified freely.
*/

#include "StdAfx.h"
#include "SampleGrab.h"
#include <mmSystem.h>

//------------------------------------------------------------------------------
//constants and defines
//------------------------------------------------------------------------------
#define RET_OK       1
#define RET_ERR      0
#define MAXPORTS    20
#define MAJORVER     1
#define MINORVER     0

//------------------------------------------------------------------------------
//                          START OF SAMPLE CODE
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//Sample code implementation - camera names and mutex protection
//------------------------------------------------------------------------------
static char* CAMERANAMES[MAXPORTS]={
	"SampleCamera-000000",
	"SampleCamera-000001",
	"SampleCamera-000002",
	"SampleCamera-000003",
	"SampleCamera-000004",
	"SampleCamera-000005",
	"SampleCamera-000006",
	"SampleCamera-000007",
	"SampleCamera-000008",
	"SampleCamera-000009",
	"SampleCamera-000010",
	"SampleCamera-000011",
	"SampleCamera-000012",
	"SampleCamera-000013",
	"SampleCamera-000014",
	"SampleCamera-000015",
	"SampleCamera-000016"
	"SampleCamera-000017"
	"SampleCamera-000018"
	"SampleCamera-000019"
};

static char* MUXNAMES[MAXPORTS]={
  "TORDIVEL AS : SAMPLEGRAB MUX.0.00",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.01",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.02",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.03",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.04",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.05",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.06",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.07",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.08",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.09",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.10",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.11",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.12",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.13",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.14",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.15",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.16",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.17",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.18",
  "TORDIVEL AS : SAMPLEGRAB MUX.0.19"
};

//------------------------------------------------------------------------------
//camera encapsulation
//------------------------------------------------------------------------------
class CCamera {
public:
	CCamera(LONG port, LPCSTR name);

  bool Open();
  bool Close();

  void LoadConfig(LPCSTR path);
  void SaveConfig(LPCSTR path);

  bool GetRange(LPCSTR prop, long  *min, long *max);
	bool GetProp(LPCSTR prop, long *value);
	bool SetProp(LPCSTR prop, long value);

	bool Grab();

  PBYTE  GetRow(int row);

	int							 m_port;				//accosiated port no
	char						 m_name[128];		//the unique camera name
	BOOL             m_open;				//flag for open/close
  UINT             m_width;				//current image width
  UINT             m_height;			//current image width
  UINT             m_pixelsize;		//current pixelsize
	PBYTE            m_buf;					//the raw buffer
	UINT         		 m_bufsize;			//size of raw buffer
	
	MMRESULT         m_timer;				//handle of grab timerfunc
	CallbackFunc    *m_oncomplete;	//callback when image done
};


//-------------------------------------------------------------------------------------------
//GLOBALS
//-------------------------------------------------------------------------------------------
static DebugMsgFunc  *debugMsgFunc = 0;						//debugmessage function pointer
static HANDLE				  mux[MAXPORTS];							//protection for open cameras
static char           szSystemPath[256] = "";			//working directory
static LONG           refcnt = 0;									//dll reference counting
static LONG    			  nCameras = 0;								//total no of available cameras
static char          *lpszCameraNames[MAXPORTS];	//name of cameras
static CCamera       *cameras[MAXPORTS];					//cameras instances
static CCamera       *xIndex[MAXPORTS];						//camera portNo cross index mapping

//-------------------------------------------------------------------------------------------
//DEBUG message routing
//-------------------------------------------------------------------------------------------
void Log(char *format, ...)
{
	if (debugMsgFunc) {
    static TCHAR   szMessage[512];
    va_list VaList;
    va_start( VaList, format);
    vsprintf_s(szMessage, format, VaList);
    debugMsgFunc(szMessage);
    va_end( VaList );
	}
}

//------------------------------------------------------------------------------
//CCamera implementation
//------------------------------------------------------------------------------
CCamera::CCamera(LONG port, LPCSTR name)
{
  m_port=port;
	strncpy_s(m_name,name,sizeof(m_name)-1);
	m_width=0;
	m_height=0;
	m_bufsize=0;
	m_buf=NULL;
	m_open=false;
	m_timer=NULL;
}

void CCamera::LoadConfig(LPCSTR path)
{
  CString fname=path;
  if (!fname.IsEmpty() && (fname.Right(1)!='\\')) fname+='\\';
  fname+=m_name;
  fname+=".cfg";
}

void CCamera::SaveConfig(LPCSTR path)
{
  CString fname=path;
  if (!fname.IsEmpty() && (fname.Right(1)!='\\')) fname+='\\';
  fname+=m_name;
  fname+=".cfg";
}

bool CCamera::Open()
{
	if (!m_open) {
		m_open=true;
		m_width=640;
		m_height=480;
		m_pixelsize=24;
		m_bufsize=m_width*m_height*m_pixelsize/8;
		m_buf=(PBYTE)malloc(m_bufsize);
		return true;
	}
	return false;
}

bool CCamera::Close()
{
	if (m_open) {
		if (m_timer) timeKillEvent(m_timer);
		m_timer=NULL;
		m_open=false;
		delete [] m_buf;
		m_buf=NULL;
		m_bufsize=0;
		return true;
	}
	return false;
}


void CALLBACK mmTimeGrabTimeout(UINT uTimerID, UINT uMsg, DWORD_PTR dwUser, DWORD_PTR dw1, DWORD_PTR dw2)
{
	CCamera* cam=(CCamera*)dwUser;
	if (cam) {
		memset(cam->m_buf,(uTimerID*3)&0xff,cam->m_bufsize);
		if (cam->m_oncomplete)
			cam->m_oncomplete(0,cam->m_port);
		cam->m_timer=NULL;
	}
}

bool CCamera::Grab()
{
	if (!m_timer) {
		m_timer=timeSetEvent(25,10,&mmTimeGrabTimeout,UINT(this),TIME_ONESHOT);
		return true;
	}
	return false;
}

PBYTE CCamera::GetRow(int row)
{
	if ((UINT)row<m_width) {
		int offset=row*m_width*m_pixelsize/8;
		return &m_buf[offset];
	}
	return NULL;
}

bool CCamera::GetRange(LPCSTR prop, long  *min, long *max)
{
	return false;
}

bool CCamera::GetProp(LPCSTR prop, long *value)
{
	return false;
}

bool CCamera::SetProp(LPCSTR prop, long value)
{
	return false;
}

//------------------------------------------------------------------------------
// Misc helper routines
//------------------------------------------------------------------------------
int GetCamIdx(LPCSTR ACamName)
{
	for (int i0=0;i0<MAXPORTS;i0++)
		if (strncmp(lpszCameraNames[i0],ACamName,sizeof(cameras[i0]->m_name))==0)
			return i0;
	return -1;
}

void QueryOpenCamera(LPCSTR ACamName, CCamera **camera)
{
  *camera=NULL;
  for (int i0=0; i0<MAXPORTS; i0++)
    if (xIndex[i0])
      if (strncmp(xIndex[i0]->m_name,ACamName,sizeof(cameras[i0]->m_name))==0) {
        *camera=xIndex[i0];
        break;
      }
}

//------------------------------------------------------------------------------
//                           END OF SAMPLE CODE
//------------------------------------------------------------------------------



//------------------------------------------------------------------------------
//
// Scorpion Camera Interface, SCI, Exported methods
//
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
//Message routing from camera driver to SCI console window
//
//Optional - not required (but very useful) 
//------------------------------------------------------------------------------
GRABDLL_API void setDebugMsgProc(DebugMsgFunc* aDebugMsgFunc)
{
	debugMsgFunc=aDebugMsgFunc;
}

//------------------------------------------------------------------------------
//Set system path for configuraton files. The camera dll is responsible for
//camera property persistance and should store camera configuration at supplied
//systempath to handle multiple configurations in different directories.
//
//Scorpion will always set the systempath to the profile directory. The driver
//may use any fileformat/subdirectories to store the individual camera
//configuratons. The configuration files should be used when opening a camera.
//
//Optional - not required (but very useful) 
//------------------------------------------------------------------------------
GRABDLL_API int setSystemPath(char *aSystemPath)
{
	strncpy_s(szSystemPath,aSystemPath,sizeof(szSystemPath)-1);
	return RET_OK;
}

//------------------------------------------------------------------------------
//Cameradriver initialization - called by Scorpion after DLL loads. Reference
//counting is strongly recommended in case of running multiple Scorpion instances
//where each Scorpion instance will call initialize/finalize at load/unload of
//the cameradriver. Initialization may also be done in dllmain but this sometimes
//causes errors in case of 
//
//Optional - not required (but very useful) 
//------------------------------------------------------------------------------
GRABDLL_API int initialize(char* aSystemPath, DebugMsgFunc* aDebugMsgFunc)
{
	setSystemPath(aSystemPath);
	setDebugMsgProc(aDebugMsgFunc);
	if (!refcnt) {
		nCameras=4;
		for (int i0=0;i0<MAXPORTS;i0++) {
			if (i0<nCameras) lpszCameraNames[i0]=CAMERANAMES[i0];
			else             lpszCameraNames[i0]=CAMERANAMES[i0];
		}
	}
	refcnt++;
	Log(_T("SampleGrab V%d.%d\n"),MAJORVER,MINORVER);
	return RET_OK;
}

GRABDLL_API int finalize()
{
	refcnt--;
	if (refcnt) nCameras=0;
	return RET_OK;
}

//------------------------------------------------------------------------------
//Camera configuration
//
//the "configure" method configures camera by grabber/port
//the "configureCamera" method configures camera by name (preferable)
//
//Scorpion uses "configureCamera" only if both methods are implemented.
//
//NOTE! 
// Scorpion versions prior to 6.1.0.370 requires the "configure" method to be
// implemented to enable configuration. From version 6.1.0.370 only "configureCamera"
// is needed.
//
//Optional - not required, but needed for configuration from Scorpion GUI
//------------------------------------------------------------------------------
#ifdef CONFIGURE_BY_PORT
GRABDLL_API int configure(HWND hParentWnd, unsigned int grabberNo, unsigned portNo)
{
	if (portNo>=MAXPORTS) return RET_ERR;
	char s0[512];
	sprintf_s(s0,_T("Configure Port %d_%d %s"),grabberNo,portNo,CAMERANAMES[portNo]);
	return ::MessageBox(hParentWnd,s0,_T("SAMPLEGRABBER CONFIGURATION"),MB_ICONEXCLAMATION | MB_OKCANCEL | MB_TASKMODAL)==IDOK ? RET_OK : RET_ERR;
}
#else
GRABDLL_API int configureCamera(HWND hParentWnd, char *cameraName)
{
	int idx=GetCamIdx(cameraName);
	if (idx<0) return RET_ERR;
	char s0[512];
	sprintf_s(s0,_T("Configure Camera %s"),CAMERANAMES[idx]);
	return ::MessageBox(hParentWnd,s0,_T("CAMERA CONFIGURATION"),MB_ICONEXCLAMATION | MB_OKCANCEL)==IDOK ? RET_OK : RET_ERR;
}
#endif

//------------------------------------------------------------------------------
//Get no of adressable grabbercards - SCI defaults to 1 if omitted
//
//Optional - rarely used, hw dependant
//------------------------------------------------------------------------------
GRABDLL_API int getCVLMaxNoOfGrabbers(void)
{
  return 1;
}

//------------------------------------------------------------------------------
//Get max no of possible concurrent cameras available. SCI defaults to 1 if not
//supported - ie. only one camera is concurrentlty supported
//
//Optional - but needed for multi camera driver implentation
//------------------------------------------------------------------------------
GRABDLL_API int getCVLMaxNoOfPorts(unsigned int grabberNo)
{
  return MAXPORTS;
}

//------------------------------------------------------------------------------
//Get no of available cameras. This method is used when iterating for avilable
//cameras in conjuction with the "getCVLCameraConfigNames" method
//
//NOTE! 
// return value should always return no of cameranames returned by
// "getCVLCameraConfigNames"
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API int getCVLNoOfSupportedCameras(void)
{
  return nCameras;
}

//------------------------------------------------------------------------------
//Get no of available cameranames. Returns pointers to 'static' strings.
//Scorpion will scan the returned **char until a NULL pointer when browsing
//available cameras.
//
//NOTE! 
// The last valid item must be followed by a NULL pointer (lpszCameraNames[nCameras]=NULL)
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API char **getCVLCameraConfigNames()
{
	return lpszCameraNames;
}


//------------------------------------------------------------------------------
//Open a named camera. The driver should return the image size on succeed, else
//(0,0). The image size is "static" as long as the camera is open and may NOT be
//changed at runtime. Changes of image size will not be 'detected' until next
//time Scorpion opens the camera. On succeed Scorpion uses the grabberNo/portNo
//as crossindex/handle for accessing the camera.
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API int openCVLPort(unsigned int grabberNo, unsigned int portNo, char* cameraName, WH &imageSize)
{
	Log(_T("Open %s\n"),cameraName);
	if (xIndex[portNo]) return RET_ERR;													//portNo busy
	int idx=GetCamIdx(cameraName);															//lookup camera
	if (idx<0) return RET_ERR;																	//undefined camera

  //check if alredy open by this instance or not
  HANDLE hMux=OpenMutex(MUTEX_ALL_ACCESS,false,MUXNAMES[portNo]);
  if (hMux!=NULL) {
    CloseHandle(hMux);																				//close it again to decrement mutex counter
		Log("Open %s - Port %d is in use\n",lpszCameraNames[idx],portNo);
		return RET_ERR;
	}
	CCamera* camera=new CCamera(portNo,lpszCameraNames[idx]);		//create the camera instance
	if (camera->Open()) {																				//open the camera, increment reference count
		xIndex[portNo]=camera;																		//set up portNo mapping
    imageSize.w=camera->m_width;  														//return size of image
		imageSize.h=camera->m_height;
		if (!mux[portNo]) mux[portNo]=CreateMutex(NULL, false, MUXNAMES[portNo]);
		camera->LoadConfig(szSystemPath);													//load configuration from configfile at systemPath directory
		return RET_OK;			
	}
	return RET_ERR;
}

//------------------------------------------------------------------------------
//Close a camera
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API void closeCVLPort(unsigned int grabberNo, unsigned int portNo)
{
  if (!xIndex[portNo]) return;
  try{
	  xIndex[portNo]->Close();
  }catch(...){
  }
	xIndex[portNo]=NULL;							//portNo is now available
	CloseHandle(mux[portNo]);					//free the resource (sample code)
	mux[portNo]=NULL;
}

//------------------------------------------------------------------------------
//grab an image from camera mapped to grabberNo/portNo
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API int doCVLGrab(unsigned int grabberNo, unsigned int portNo)
{
  try{
	  if (!xIndex[portNo]) return RET_ERR;
	  CCamera *camera=xIndex[portNo];
		return camera->Grab() ? RET_OK : RET_ERR;
  }catch(...){
  	return RET_ERR;
	}
}

//------------------------------------------------------------------------------
//get teh image from camera mapped to grabberNo/portNo
// rowPtr   - [IN] destination buffer where to copy image date - allocated/owned by Scorpion
// row      - [IN] range 0..height-1 image layout TOP-DOWN
// firstPix - [IN] index of first pixel in row (normally 0)
// lastPix  - [IN] index of last pixel in row (normally width-1)
//
//NOTE!
// since the image size is "static" at runtime, Scorpion requires image size returned in
// the open method. The pixelSize may be changed at runtime if the cameradriver supports
// "set/getProperty" methods and returns current "PixelSize", see the "getProperty" method.
// Scorpion always checks "PixelSize" property if available before aqcuring an image.
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API int getCVLGrabRow(unsigned int grabberNo, unsigned int portNo, unsigned char* rowPtr, unsigned int row, unsigned int firstPix, unsigned int lastPix)
{
  try{
	  if (!xIndex[portNo]) return RET_ERR;
	  CCamera *camera=xIndex[portNo];
	  PBYTE pBuf=camera->GetRow(row);
		if (!pBuf) return RET_ERR;																					//invalid row
		memcpy(rowPtr,pBuf,(lastPix-firstPix+1)*(camera->m_pixelsize/8));		//copy image data, note that subrow may be requested
	  return RET_OK;
  }catch(...){
  }
  return RET_ERR;
}

//------------------------------------------------------------------------------
//setup camera for hw trigged control at runtime. Normally trigger options is set
//up "static" in the configuration dialog.
//
//Optional - rarely used
//------------------------------------------------------------------------------
GRABDLL_API int setHWTrigger(unsigned int grabberNo, unsigned int portNo, unsigned int HWTrigger)
{
  try{
 	  if (!xIndex[portNo]) return RET_ERR;
    return RET_ERR;
  }catch(...){
    return RET_ERR;
  }
}

//------------------------------------------------------------------------------
//set a callback for image complete notifications. The callback is normally
//called from the camera thread and only gives a notification to Scorpion a new
//image is completed. The callback method is thread safe.
//
//Required
//------------------------------------------------------------------------------
GRABDLL_API int setCompleteCallback(unsigned int grabberNo, unsigned int portNo, CallbackFunc* func)
{
	try {
 		if (!xIndex[portNo]) return RET_ERR;
		xIndex[portNo]->m_oncomplete=func;
		return RET_OK;
  }catch(...){
    return RET_ERR;
  }
}

//------------------------------------------------------------------------------
//get range for a named camera property
// prop - [IN]  name of property
// min  - [OUT] min value of property
// max  - [OUT] max value of property
//
//Optional
//------------------------------------------------------------------------------
GRABDLL_API int getPropertyRange(unsigned int grabberNo, unsigned int portNo, char *prop, long *min, long *max)
{
  try{
 	  if (!xIndex[portNo]) return RET_ERR;
    return xIndex[portNo]->GetRange(prop,min,max) ? RET_OK : RET_ERR;
  }catch(...){
    return RET_ERR;
  }
}

//------------------------------------------------------------------------------
//get current value for a named camera property
// prop  - [IN]  name of property
// value - [OUT] value of property
//
//NOTE!
// special case for PixelSize since the open method don't report PixelSize.
// Pixelsize must be either 8, 24 or 32. if not supported SCI defaults to 8
// ie. only grey scale images are used by SCI
//
//Optional - required if driver supports color images
//------------------------------------------------------------------------------
GRABDLL_API int getProperty(unsigned int grabberNo, unsigned int portNo, char *prop, long *value)
{
  try{
 	  if (!xIndex[portNo]) return RET_ERR;
		if (!_stricmp("PixelSize",prop)) {
			*value=xIndex[portNo]->m_pixelsize;
			return RET_OK;
		}
    return xIndex[portNo]->GetProp(prop,value) ? RET_OK : RET_ERR;
  }catch(...){
    return RET_ERR;
  }
}

//------------------------------------------------------------------------------
//get current value for a named camera property
// prop  - [IN]  name of property
// value - [OUT] value of property
//
//Optional
//------------------------------------------------------------------------------
GRABDLL_API int setProperty(unsigned int grabberNo, unsigned int portNo, char *prop, long value)
{
  try{
 	  if (!xIndex[portNo]) return RET_ERR;
    return xIndex[portNo]->SetProp(prop,value) ? RET_OK : RET_ERR;
  }catch(...){
    return RET_ERR;
  }
}

//------------------------------------------------------------------------------
//set command callback function
// func  - [IN]  cmd callback
//
//Optional
//------------------------------------------------------------------------------
GRABDLL_API int setCmdCallback(unsigned int grabberNo, unsigned int portNo, CmdCallbackFunc* func)
{
  try{
 	  if (!xIndex[portNo]) return RET_ERR;
    return RET_OK;
  }catch(...){
    return RET_ERR;
  }
}

//------------------------------------------------------------------------------
//execute command
// cmd    - [IN] name of cmd
// params - [IN] command params
// resp   - [OUT] command response
// size   - [IN] max size of response string
//
//Optional
//------------------------------------------------------------------------------
GRABDLL_API int execCmd(unsigned int grabberNo, unsigned int portNo, const char* cmd, const char* params, char* resp, int size)
{
  try{
 	  if (!xIndex[portNo]) return RET_ERR;
		if (!_stricmp("get",cmd)) {
      if (!_stricmp("name",params)) 
        strcpy_s(resp,size-1,xIndex[portNo]->m_name);
      else 
        strcpy_s(resp,size-1,"unknowncmd");
			return RET_OK;
		}
  }catch(...){
    return RET_ERR;
  }
}
